from typing import Optional, Union, Tuple
import os

import h5py
import numpy as np

from ._make import Make
from ._ovf import save_ovf, load_ovf, get_ovf_parms
from ._plot import imshow
from ._disp import Disp


class Llyr:
    def __init__(
        self, h5_path: str, load_path: Optional[str] = None, tmax=None, force=False
    ) -> None:
        self.h5_path = self._clean_path(h5_path)
        self.name = self.h5_path.split("/")[-1][:-3]
        if load_path is not None:
            Make(self, load_path, force=force, tmax=tmax)
        self._getitem_dset: Optional[str] = None

    def __repr__(self):
        return f"Llyr('{self.h5_path}')"

    def __str__(self):
        return f"Llyr('{self.h5_path}')"

    def _clean_path(self, path):
        path = os.path.abspath(path)
        if path[-3:] != ".h5":
            out_path = f"{path}.h5"
        else:
            out_path = path
        return out_path

    def __getitem__(
        self,
        index: Union[str, Tuple[Union[int, slice], ...]],
    ) -> Union["Llyr", float, np.ndarray]:
        with h5py.File(self.h5_path, "r") as f:

            # if slicing

            if isinstance(index, (slice, tuple, int)):
                # if dset is defined
                if isinstance(self._getitem_dset, str):
                    out_dset: np.ndarray = f[self._getitem_dset][index]
                    self._getitem_dset = None
                    return out_dset
                else:
                    raise AttributeError("You can only slice datasets")

            elif isinstance(index, str):
                # if dataset
                if index in list(f.keys()):
                    self._getitem_dset = index
                    return self
                # if attribute
                elif index in list(f.attrs.keys()):
                    out_attribute: float = f.attrs[index]
                    return out_attribute
                else:
                    raise KeyError("No such Dataset or Attribute")
            else:
                raise TypeError()

    def set_attr(
        self,
        name: str,
        key: str,
        val: Union[str, int, float, slice, Tuple[Union[int, slice], ...]],
    ) -> None:
        """set a new attribute"""
        with h5py.File(self.h5_path, "a") as f:
            f[name].attrs[key] = val

    def shape(self, dset: str) -> (int, int, int, int, int):
        with h5py.File(self.h5_path, "r") as f:
            return f[dset].shape

    def save_mx3(self, savepath: str) -> None:
        """prints or saves the mx3"""
        with open(savepath, "w") as f:
            f.writelines(self["mx3"])

    def save_ovf(self, ovf_path: str, dset: str, t: int = 0) -> None:
        arr = self[dset][t]
        save_ovf(ovf_path, arr, self.dx, self.dy, self.dz)

    def imshow(self, dset, t=-1, zero=True):
        imshow(self, dset, t=t, zero=zero)

    @staticmethod
    def load_ovf(ovf_path: str):
        return load_ovf(ovf_path)

    @property
    def mx3(self) -> str:
        print(self["mx3"])

    @property
    def dt(self) -> float:
        return self["dt"]

    @property
    def dx(self) -> float:
        return self["dx"]

    @property
    def dy(self) -> float:
        return self["dy"]

    @property
    def dz(self) -> float:
        return self["dz"]

    @property
    def p(self) -> None:
        with h5py.File(self.h5_path, "r") as f:
            print("Datasets:")
            for key, val in f.items():
                print(f"    {key:<15}: {val.shape}")
                if f[key].attrs:
                    print(f"    Attributes of {key}:")
                for akey, aval in f[key].attrs.items():
                    if isinstance(aval, np.ndarray):
                        aval = f"{aval.shape} : min={aval.min()}, max={aval.max()}"
                    print(f"        {akey:<11}= {aval}")

            print("Global Attributes:")
            for key, val in f.attrs.items():
                if key in ["mx3", "script"]:
                    val = val.replace("\n", "")
                    print(f"    {key:<15}= {val[:10]}...")
                else:
                    print(f"    {key:<15}= {val}")

    @property
    def dsets(self) -> list:
        with h5py.File(self.h5_path, "r") as f:
            dsets = list(f.keys())
        return dsets

    @property
    def attrs(self) -> list:
        with h5py.File(self.h5_path, "r") as f:
            attrs = list(f.attrs.keys())
        return attrs

    def delete(self, dset: str) -> None:
        """deletes dataset"""
        with h5py.File(self.h5_path, "a") as f:
            del f[dset]

    def move(self, source: str, destination: str) -> None:
        """move dataset or attribute"""
        with h5py.File(self.h5_path, "a") as f:
            f.move(source, destination)



from typing import Tuple, Union
import multiprocessing as mp

import h5py
import dask.array as da
from tqdm.notebook import tqdm
import numpy as np
import psutil
from dask.distributed import Client


class Disp:
    def __init__(self, h5):
        self.h5 = h5

    def disp(
        self,
        dset: str = "WG",
        name: str = "disp",
        slices: Tuple[Union[int, slice], ...] = (
            slice(None),
            slice(None),
            slice(None),
            slice(None),
            2,
        ),
        save: bool = True,
        force: bool = False,
    ) -> None:
        """Calculates and returns the dispersions using dask"""
        if name in self.h5.dsets():
            if force:
                with h5py.File(self.h5.h5_path, "a") as f:
                    del f[name]
            else:
                input_string: str = input(
                    f"{name} is already a dataset, [y] to overwrite, [n] to cancel, else [input] a new name"
                )
                if input_string.lower() == "y":
                    with h5py.File(self.h5.h5_path, "a") as f:
                        del f[name]
                elif input_string.lower() == "n":
                    return
                else:
                    name = input_string
        dask_client = self.start_dask_client()
        with h5py.File(self.h5.h5_path, "r") as f:
            arr = da.from_array(f[dset], chunks=(None, None, 15, None, None))
            arr = arr[slices]  # slice
            arr = da.multiply(
                arr, np.hanning(arr.shape[0])[:, None, None, None]
            )  # hann filter on the t axis
            arr = arr.sum(axis=1)  # t,z,y,x => t,y,x sum of z
            arr = da.moveaxis(arr, 1, 0)  # t,y,x => y,t,x swap t and y
            ham2d = np.sqrt(
                np.outer(np.hanning(arr.shape[1]), np.hanning(arr.shape[2]))
            )  # shape(t, x)
            arr = da.multiply(arr, ham2d[None, :, :])  # hann window on t and x
            arr = da.fft.fft2(arr)  # 2d fft on t and x
            arr = da.subtract(
                arr, da.average(arr, axis=(1, 2))[:, None, None]
            )  # substract the avr of t,x for a given y
            arr = da.moveaxis(arr, 0, 1)
            arr = arr[: arr.shape[0] // 2]  # split f in 2, take 1st half
            arr = da.fft.fftshift(arr, axes=(1, 2))
            arr = da.absolute(arr)  # from complex to real
            arr = da.sum(arr, axis=1)  # sum y
            out = arr.compute()
        dask_client.close()

        if save:
            with h5py.File(self.h5.h5_path, "a") as f:
                dset_disp = f.create_dataset(name, data=out)
                dset_disp.attrs["slices"] = str(slices)
                dset_disp.attrs["dset"] = dset

    def fft(
        self,
        dset: str = "ND",
        name: str = "fft",
        slices: Tuple[Union[int, slice], ...] = (
            slice(None),
            slice(None),
            slice(None),
            slice(None),
            2,
        ),
        save: bool = True,
    ) -> np.ndarray:
        """Calculates and return the fft of the dataset"""
        with h5py.File(self.h5.h5_path, "a") as f:
            if slices is None:
                arr = f[dset][:]
            else:
                arr = f[dset][slices]
            arr -= arr[0]

            for i in [0, 2, 3]:
                if arr.shape[i] % 2 == 0:
                    arr = np.delete(arr, 1, i)

            hann = np.hanning(arr.shape[0])
            for i, _ in enumerate(hann):
                arr[i] *= hann[i]

            arr = arr.sum(axis=1)  # t,z,y,x,c => t,y,x,c
            fft = []  # fft for each cell and comp
            for y in tqdm(
                range(arr.shape[1]),
                desc="Calculating FFT",
                total=arr.shape[1],
                leave=False,
            ):  # y
                for x in range(arr.shape[2]):  # x
                    for c in range(arr.shape[3]):
                        d = arr[:, y, x, c]
                        d = d - np.average(d)
                        fft.append(np.fft.rfft(d))
            out = np.array(fft)
            out = np.abs(out)
            out = np.sum(out, axis=0)
            out /= (
                arr.shape[1] * arr.shape[2]
            )  # changing the amplitude on a per cell basis

            if save:
                dset_fft = f.create_dataset(name, data=out)
                dset_fft.attrs["slices"] = str(slices)
                dset_fft.attrs["dset"] = dset
        return out

    @property
    def freqs(self) -> np.ndarray:
        """returns frequencies in GHz depending on the number of t points in the dset and value of dt"""
        with h5py.File(self.h5.h5_path, "r") as f:
            freqs = np.fft.rfftfreq(f["disp"].shape[0], self.h5.dt * 1e9)

        return freqs

    @property
    def kvecs(self) -> np.ndarray:
        """returns wavevectors"""
        with h5py.File(self.h5.h5_path, "r") as f:
            kvecs = (
                np.fft.fftshift(
                    np.fft.fftfreq(f["disp"].shape[1], self.h5.dx) * 2 * np.pi
                )
                * 1e-6
            )
        return kvecs

    def start_dask_client(self, port=23232):
        ram = int(psutil.virtual_memory().free / 1e9 * 0.95)
        print(f"Dask client started at 127.0.0.1:{port} with {ram} GB of ram")
        client = Client(
            processes=False,
            threads_per_worker=mp.cpu_count(),
            n_workers=1,
            memory_limit=f"{ram} GB",
        )
        return client



import matplotlib.pyplot as plt
import cmocean  # pylint: disable=unused-import


def imshow(self, dset: str, zero: bool = True, t: int = -1, c: int = 2):
    fig, ax = plt.subplots(1, 1, dpi=200)
    if zero:
        arr = self[dset][[0, t], 0, :, :, c]
        arr = arr[1] - arr[0]
    else:
        arr = self[dset][t, 0, :, :, c]
    amin, amax = arr.min(), arr.max()
    if amin < 0 and amax > 0:
        cmap = "cmo.balance"
        vmm = max((-amin, amax))
        vmin, vmax = -vmm, vmm
    else:
        cmap = "cmo.amp"
        vmin, vmax = amin, amax

    ax.imshow(
        arr,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        extent=[0, arr.shape[1] * self.dx * 1e9, 0, arr.shape[0] * self.dy * 1e9],
    )
    ax.set(
        title=self.name,
        xlabel="x (nm)",
        ylabel="y (nm)",
    )
    fig.colorbar(ax.get_images()[0], ax=ax)

    return ax



from typing import Tuple, List, Optional
from glob import glob
import os
import re
import multiprocessing as mp

import h5py
import numpy as np
from tqdm import tqdm

from ._ovf import load_ovf, get_ovf_parms


class Make:
    def __init__(self, h5, load_path, force=False, tmax=None):
        self.h5 = h5
        self.make(load_path, force, tmax)

    def add_table(self, table_path: str, dset_name: str = "table") -> None:
        """Adds a the mumax table.txt file as a dataset"""
        if os.path.isfile(table_path):
            with open(table_path, "r") as table:
                header: str = table.readline()
                data: np.ndarray = np.loadtxt(table).T
                dt: float = (data[0, -1] - data[0, 0]) / (data.shape[1] - 1)
            with h5py.File(self.h5.h5_path, "a") as f:
                tableds = f.create_dataset(dset_name, data=data)
                tableds.attrs["header"] = header
                f.attrs["dt"] = dt

    def _get_paths(self, load_path: str) -> Tuple[str, str]:
        """Cleans the input string and return the path for .out folder and .mx3 file"""
        load_path = load_path.replace(".mx3", "").replace(".out", "").replace(".h5", "")
        out_path = f"{load_path}.out"
        mx3_path = f"{load_path}.mx3"
        return out_path, mx3_path

    def add_mx3(self, mx3_path: str) -> None:
        """Adds the mx3 file to the f.attrs"""
        if os.path.isfile(mx3_path):
            with open(mx3_path, "r") as mx3:
                with h5py.File(self.h5.h5_path, "a") as f:
                    f.attrs["mx3"] = mx3.read()
        else:
            print(f"{mx3_path} not found")

    def _create_h5(self, force) -> bool:
        """Creates an empty .h5 file"""
        if force:
            with h5py.File(self.h5.h5_path, "w"):
                return True
        else:
            if os.path.isfile(self.h5.h5_path):
                input_string: str = input(
                    f"{self.h5.h5_path} already exists, overwrite it [y/n]?"
                )
                if input_string.lower() in ["y", "yes"]:
                    with h5py.File(self.h5.h5_path, "w"):
                        return True
        return False

    def _get_dset_prefixes(self, out_path: str) -> List[str]:
        """From the .out folder, get the list of prefixes, each will correspond to a different dataset"""
        paths = glob(f"{out_path}/*.ovf")
        prefixes = {re.sub(r"_?[\d.]*.ovf", "", path.split("/")[-1]) for path in paths}
        return list(prefixes)

    def _get_dset_name(self, prefix: str) -> str:
        """From the prefix, this tries to return a human readable version"""
        common_prefix_to_name = (
            ("m_xrange", "ND"),
            ("m_zrange", "WG"),
            ("B_demag_xrange", "ND_B"),
            ("B_demag_zrange", "WG_B"),
            # ("stable", "stable"),
        )
        for i in common_prefix_to_name:
            if i[0] in prefix:
                return i[1]
        return prefix

    def _save_stepsize(self, parms: dict) -> None:
        for key in ["dx", "dy", "dz"]:
            if key not in self.h5.attrs:
                with h5py.File(self.h5.h5_path, "a") as f:
                    f.attrs[key] = parms[key]

    def add_dset(
        self,
        out_path: str,
        prefix: str,
        name: Optional[str] = None,
        tmax: Optional[int] = None,
        force: bool = False,
    ) -> None:
        """Creates a dataset from an input .out folder path and a prefix (i.e. "m00")"""
        ovf_paths = sorted(glob(f"{out_path}/{prefix}*.ovf"))[:tmax]
        # load one file to initialize the h5 dataset with the correct shape
        ovf_parms = get_ovf_parms(ovf_paths[0])
        self._save_stepsize(ovf_parms)
        ovf_shape = ovf_parms["shape"]
        dset_shape = (len(ovf_paths),) + ovf_shape
        if name is None:
            name = self._get_dset_name(prefix)

        with h5py.File(self.h5.h5_path, "a") as f:
            if force and name in list(f.keys()):
                del f[name]
            dset = f.create_dataset(name, dset_shape, np.float32)
            with mp.Pool(processes=int(mp.cpu_count())) as p:
                for i, data in enumerate(
                    tqdm(
                        p.imap(load_ovf, ovf_paths),
                        leave=False,
                        desc=name,
                        total=len(ovf_paths),
                    )
                ):
                    dset[i] = data

    def add_np_dset(self, arr: np.ndarray, name: str, force: bool = False):
        with h5py.File(self.h5.h5_path, "a") as f:
            if name in self.h5.dsets():
                if force:
                    del f[name]
                    f.create_dataset(name, data=arr)
                else:
                    raise NameError(
                        "A dataset with this name already exists, use force=True to override it"
                    )
            else:
                f.create_dataset(name, data=arr)

    def make(self, load_path: str, force: bool, tmax: Optional[int]) -> None:
        """Automatically parse the load_path and will create datasets"""
        self._create_h5(force)
        out_path, mx3_path = self._get_paths(load_path)
        self.add_table(f"{out_path}/table.txt")
        self.add_mx3(mx3_path)
        dset_prefixes = self._get_dset_prefixes(out_path)
        for dset_prefix in dset_prefixes:
            self.add_dset(out_path, dset_prefix, tmax=tmax)
